//Copyright (C) 2004, Brian Enigma <enigma at netninja.com>
//This file is part of Tasty Safari.
//
//Tasty Safari is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//Tasty Safari is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with Foobar; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package org.ninjasoft.tastysafari.logic;

import java.net.*;
import java.io.*;
import java.util.*;

import org.ninjasoft.tastysafari.data.*;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import org.apache.xpath.XPathAPI;

/**
 * Java API for accessing the Del.icio.us HTTP/XML API
 * @author enigma
 */
public class DeliciousApi {
    private Authenticator authenticator = null;
    private String server;
    private String username;
    
    public DeliciousApi(String username, String password) {
        this(username, password, "del.icio.us");
    }
    
    public DeliciousApi(String username, String password, String server) {
        this.authenticator = new HttpHeadlessAuthenticator(username, password);
        this.server = server;
        this.username = username;
    }
    
    /**
     * Return a list of all of the subscriptions for a given user.
     * @return a list of Subscription objects
     */
    public Subscription[] getSubs() throws MalformedURLException, SAXException, IOException, ParserConfigurationException, FactoryConfigurationError, BadHttpResponse, TransformerException {
        Document doc = makeRequestXml("/api/inbox/subs");
        Vector subs = new Vector();
        NodeList nodes = XPathAPI.selectNodeList(doc, "//sub");
        for (int i=0; i<nodes.getLength(); i++) {
            Subscription sub = new Subscription();
            sub.setUser(getAttribute(nodes.item(i), "user"));
            sub.setTags(getAttribute(nodes.item(i), "tag"));
            subs.add(sub);
        }
        int counter = 0;
        Subscription[] result = new Subscription[subs.size()];
        for (Iterator i = subs.iterator(); i.hasNext(); counter++)
            result[counter] = (Subscription) i.next();
        return result;
    }
    
    /**
     * Return a list of all dates that had entries.  The dates are of a format useful for
     * passing to the getPosts function.
     * @return an array of date strings
     */
    public String[] getDates() throws MalformedURLException, SAXException, IOException, ParserConfigurationException, FactoryConfigurationError, BadHttpResponse, TransformerException {
        return getDates("/api/posts/dates");
    }
    
    /**
     * Return a list of all dates that had entries.  The dates are of a format useful for
     * passing to the getPosts function.
     * @return an array of date strings
     */
    public String[] getInboxDates() throws MalformedURLException, SAXException, IOException, ParserConfigurationException, FactoryConfigurationError, BadHttpResponse, TransformerException {
        return getDates("/api/inbox/dates");
    }

    /**
     * Wrapper that extracts dates from <date date="xyz"> XML, given a particular API URL
     * @param apiCall Either "/api/posts/dates" or "/api/inbox/dates"
     * @return an array of String dates
     */
    private String[] getDates(String apiCall) throws MalformedURLException, SAXException, IOException, ParserConfigurationException, FactoryConfigurationError, BadHttpResponse, TransformerException {
        Document doc = makeRequestXml(apiCall);
        Vector dates = new Vector();
        NodeList nodes = XPathAPI.selectNodeList(doc, "//date");
        for (int i=0; i<nodes.getLength(); i++)
            dates.add(getAttribute(nodes.item(i), "date"));
        int counter = 0;
        String[] result = new String[dates.size()];
        for (Iterator i = dates.iterator(); i.hasNext(); counter++)
            result[counter] = i.next().toString();
        return result;
    }

    /**
     * Return a list of all of your posts on a given date.
     * @param date The date string, as returned by getDates (or null for the most recent)
     * @return an array of PostEntry objects
     */
    public PostEntry[] getPosts(String date) throws MalformedURLException, SAXException, IOException, ParserConfigurationException, FactoryConfigurationError, BadHttpResponse, TransformerException {
        return getPosts("/api/posts/get", date);
    }

    /**
     * Return a list of all inbox posts on a given date.
     * @param date The date string, as returned by getDates (or null for the most recent)
     * @return an array of PostEntry objects
     */
    public PostEntry[] getInbox(String date) throws MalformedURLException, SAXException, IOException, ParserConfigurationException, FactoryConfigurationError, BadHttpResponse, TransformerException {
        return getPosts("/api/inbox/get", date);
    }
    
    private PostEntry[] getPosts(String apiCall, String date) throws MalformedURLException, SAXException, IOException, ParserConfigurationException, FactoryConfigurationError, BadHttpResponse, TransformerException {
        Vector entries = new Vector();
        String url = apiCall;
        if (date != null)
            url = url + "?dt=" + date;
        Document doc = makeRequestXml(url);
        NodeList nodes = XPathAPI.selectNodeList(doc, "//post");
        for (int i=0; i<nodes.getLength(); i++)
            entries.add(parseEntry(nodes.item(i)));
        int counter = 0;
        PostEntry[] result = new PostEntry[entries.size()];
        for (Iterator i = entries.iterator(); i.hasNext(); counter++)
            result[counter] = (PostEntry) i.next();
        return result;
    }
    
    private PostEntry parseEntry(Node node) {
        NamedNodeMap attributes = node.getAttributes();
        PostEntry result = new PostEntry();
        result.setHref(getAttribute(node, "href"));
        result.setTags(getAttribute(node, "tags"));
        result.setDescription(getAttribute(node, "description"));
        result.setExtended(getAttribute(node, "extended"));
        result.setTime(getAttribute(node, "time"));
        result.setUser(getAttribute(node, "user"));
        return result;
    }
    
    /**
     * Given an XML node, return the value of the requested attribute (or an empty string
     * if it does not exist).
     * 
     * NOTE: this is slightly inefficient if you are going to call it multiple times for
     * the same node, as it has to call getAttributes during every function call.  This is
     * the price you pay for never getting a NullPointerException and never returning a null.
     * @param node
     * @param attributeName
     * @return
     */
    private String getAttribute(Node node, String attributeName) {
        if (node == null) return "";
        NamedNodeMap attributes = node.getAttributes();
        if (attributes == null) return "";
        Node valueNode = attributes.getNamedItem(attributeName);
        if (valueNode == null) return "";
        String value = valueNode.getNodeValue();
        if (value == null) return "";
        return value;
    }
    
    private Document makeRequestXml(String path) throws MalformedURLException, SAXException, IOException, ParserConfigurationException, FactoryConfigurationError, BadHttpResponse {
        return DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(new ByteArrayInputStream(makeRequestString(path).getBytes()));
    }
    
    private String makeRequestString(String path) throws MalformedURLException, IOException, BadHttpResponse {
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        URL url = new URL("http", this.server, 80, path); 
        Authenticator.setDefault(this.authenticator);
        HttpURLConnection.setFollowRedirects(true);
        HttpURLConnection connection = (HttpURLConnection) url.openConnection();
        if (connection.getResponseCode() != 200) {
            throw new BadHttpResponse("Expected HTTP response 200, got " + connection.getResponseCode() + ":" + connection.getResponseMessage());
        }
        InputStream in = connection.getInputStream();
        byte[] buff = new byte[1024];
        int len;
        while ((len = in.read(buff)) > 0)
            out.write(buff, 0, len);
        in.close();
        out.close();
        connection.disconnect();
        Authenticator.setDefault(null);
        // Assuming it is all 8-bit ASCII
        return clean(out.toString("UTF-8"));
    }
    
    /**
     * Clean any high-ASCII from a string
     * @param s
     * @return
     */
    private String clean(String s) {
        StringBuffer result = new StringBuffer();
        for (int i=0; i<s.length(); i++) {
            char c = s.charAt(i);
            if ((c >= 9) && (c <= '~'))
                result.append(c);
        }
        return result.toString();
    }

    /**
     * Gets the username and password from the user without a prompt (basically, by resupplying
     * the data that was passed into the constructor of the containing class).
     * @author enigma
     */
    public class HttpHeadlessAuthenticator extends Authenticator {
        private String username;
        private String password;
        public HttpHeadlessAuthenticator(String username, String password) {
            this.username = username;
            this.password = password;
        }
        protected PasswordAuthentication getPasswordAuthentication() {
            String promptString = getRequestingPrompt();
            String hostname = getRequestingHost();
            InetAddress ipaddr = getRequestingSite();
            int port = getRequestingPort();
            
            return new PasswordAuthentication(this.username, this.password.toCharArray());
        }
    }
    
    /**
     * Exception used to signify a bad response (404, 500, etc...basically anything aside from 200)
     * @author enigma
     */
    public class BadHttpResponse extends Exception {
        public BadHttpResponse(String message) {
            super(message);
        }
    }
    
    /**
     * Simple unit test function
     * @param argv
     */
    public static void main(String[] argv) throws Exception {
        if (argv.length != 2) {
            System.out.println("Place username and password on command line");
            System.out.println("(Don't run this test on a public Unix system because anyone with the \"ps\"");
            System.out.println("command will be able to see your password on the command line)");
            return;
        }
        DeliciousApi api = new DeliciousApi(argv[0], argv[1]);
        //test1(api);
        test2(api);
    }
        
    /** Show "my" entries */
    public static void test1(DeliciousApi api) throws Exception {
        String[] dates = api.getDates();
        System.out.println("Found " + dates.length + " date(s)");
        for (int i=0; i<dates.length; i++) {
            System.out.println("##### on date " + dates[i] + ":");
            PostEntry[] posts = api.getPosts(dates[i]);
            System.out.println("Found " + posts.length + " post(s)");
            for (int j=0; j<posts.length; j++)
                System.out.println(posts[j]);
        }
    }
    
    /** Show "my friends'" entries */
    public static void test2(DeliciousApi api) throws Exception {
        String[] dates = api.getInboxDates();
        System.out.println("Found " + dates.length + " date(s)");
        for (int i=0; i<dates.length; i++) {
            System.out.println("##### on date " + dates[i] + ":");
            PostEntry[] posts = api.getInbox(dates[i]);
            System.out.println("Found " + posts.length + " post(s)");
            for (int j=0; j<posts.length; j++)
                System.out.println(posts[j]);
        }
    }
}
