//Copyright (C) 2004, Brian Enigma <enigma at netninja.com>
//This file is part of Tasty Safari.
//
//Tasty Safari is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//Tasty Safari is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with Foobar; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package org.ninjasoft.tastysafari.logic;

import java.io.*;
import java.rmi.server.UID;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import org.apache.xpath.XPathAPI;
import org.apache.xml.serialize.XMLSerializer;
import org.apache.xml.serialize.OutputFormat;

import org.ninjasoft.tastysafari.data.*;

/**
 * Class for manipulating Safari bookmarks
 * @author enigma
 */
public class SafariBookmarkEditor {
    /** Where in your home directory does Safari keep its data */
    public static final String relativePath = "Library/Safari/";
    /** Which filename contains the bookmarks */
    public static final String bookmarkFilename = "Bookmarks.plist";
    /** User's home directory */
    private String homeDir = "";
    private String bookmarkFile = "";

    public SafariBookmarkEditor() {
        this.homeDir = System.getProperty("user.home");
        this.bookmarkFile = this.homeDir + "/" + relativePath + bookmarkFilename;
    }
    
    /**
     * Perform a save of the default Bookmarks.plist file.  NOTE: this does not perform a
     * backup of the existing file.  If a backup is desired, the caller should call
     * backupBookmarks.
     */
    public void save(Document doc) throws IOException {
        save(bookmarkFile, doc);
    }

    /**
     * Perform a save of the bookmarks to a user-specified file name.  This is useful for
     * debugging and unit testing without actually overwriting your default
     * Bookmarks.plist file.
     * @param filename
     */
    public void save(String filename, Document doc) throws IOException {
        FileOutputStream out;
        File f = new File(filename);
        OutputFormat format = new OutputFormat();
        format.setIndenting(true);
        format.setIndent(4);
        format.setLineWidth(1024);
        format.setPreserveEmptyAttributes(true);
        format.setPreserveSpace(true);
        XMLSerializer serializer = new XMLSerializer(format);
        out = new FileOutputStream(f);
        serializer.setOutputByteStream(out);
        serializer.serialize(doc);
        out.close();
    }
    
    /**
     * Copy Bookmarks.plist to Bookmarks.plist.1, doing a cascade rename
     * all the way down to Bookmarks.plist.9, which gets deleted and replaced
     * with Bookmarks.plist.8.
     */
    public void backupBookmarks() throws IOException {
        for (int i=8; i>=1; i--) {
            File dest = new File(bookmarkFile + "." + (i+1));
            File source = new File(bookmarkFile + "." + i);
            if (dest.exists())
                dest.delete();
            source.renameTo(dest);
        }
        FileInputStream in = new FileInputStream(new File(bookmarkFile));
        FileOutputStream out = new FileOutputStream(new File(bookmarkFile + ".1"));
        byte[] buff = new byte[1024];
        int len;
        while ((len = in.read(buff)) > 0)
            out.write(buff, 0, len);
        in.close();
        out.close();
    }
    
    /**
     * Load the Bookmarks.plist file into a DOM and return the DOM.
     * @return
     * @throws SAXException
     * @throws IOException
     * @throws ParserConfigurationException
     * @throws FactoryConfigurationError
     */
    public Document load() throws SAXException, IOException, ParserConfigurationException, FactoryConfigurationError {
        return DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(new File(bookmarkFile));
    }
    
    /**
     * Given a parent node representing a bookmark folder (an item with type WebBookmarkTypeList), remove
     * all child items, effectively turning it into an empty folder ready to be populated with 
     * links and subfolders.
     * @param parentNode
     * @throws TransformerException
     * @throws UnexpectedXmlFileFormatException
     */
    public static void deleteChildren(Node parentNode) throws TransformerException, UnexpectedXmlFileFormatException {
        // Reference to the owner document (used to quickly generate new nodes)
        Document doc = parentNode.getOwnerDocument();
        // Look for <key>Children</key> followed by <array>....</array> and remove
        Node childrenNode = XPathAPI.selectSingleNode(parentNode, "key[child::text()='Children']");
        Node arrayNode = null;
        if (childrenNode != null) {
            // Skip to the next non #text, non #comment node
            arrayNode = childrenNode.getNextSibling();
            while((arrayNode != null) && (arrayNode.getNodeName().startsWith("#"))) {
            	arrayNode = arrayNode.getNextSibling();
            }
            if (arrayNode == null)
                throw new UnexpectedXmlFileFormatException("Children node did not have a sibling");
            if (!arrayNode.getNodeName().equals("array"))
                throw new UnexpectedXmlFileFormatException("Children node had unexpected sibling (" + arrayNode.getNodeName() + " instead of array)");
            parentNode.removeChild(arrayNode);
            parentNode.removeChild(childrenNode);
        }
        // ASSERT: at this point, there is no <key>Children</key>\n<array>...</array> under the parent node
        // We can now safely create one
        childrenNode = createTag(doc, "key", "Children");
        arrayNode = doc.createElement("array");
        parentNode.appendChild(childrenNode);
        parentNode.appendChild(arrayNode);
    }
    
    /**
     * Given a parent node representing a bookmark folder (an item with type "WebBookmarkTypeList"), 
     * delete all existing children children
     * @param parentNode
     * @param username
     * @param folderName
     * @param entries
     * @throws TransformerException
     * @throws UnexpectedXmlFileFormatException
     */
    public static void graftBookmarks(Element parentNode, String username, String folderName, PostEntry[] entries) throws TransformerException, UnexpectedXmlFileFormatException {
        // Reference to the owner document (used to quickly generate new nodes)
        Document doc = parentNode.getOwnerDocument();
        // Look for existing "<key>Children</key>\n<array>...." node set 
        Node arrayNode = getChildrenArray(parentNode);
        // Now that we have the arrayNode, we can add child folders
        Element usernameFolder = locateCreateSubFolder(parentNode, username);
        Node usernameFolderArray = getChildrenArray(usernameFolder);
        Element categoryFolder = createFolderNode(doc, folderName);
        Node categoryFolderArray = getChildrenArray(categoryFolder);
        usernameFolderArray.appendChild(categoryFolder);
        for (int i=0; i<entries.length; i++)
            categoryFolderArray.appendChild(createBookmarkNode(doc, entries[i]));
        //arrayNode.appendChild(usernameFolder);
    }
    
    public static void addLink(Element parentNode, String name, String url) throws TransformerException, UnexpectedXmlFileFormatException {
        // Reference to the owner document (used to quickly generate new nodes)
        Document doc = parentNode.getOwnerDocument();
        // Look for existing "<key>Children</key>\n<array>...." node set 
        Node arrayNode = getChildrenArray(parentNode);
        PostEntry entry = new PostEntry();
        entry.setHref(url);
        entry.setDescription(name);
        arrayNode.appendChild(createBookmarkNode(doc, entry));
    }
    
    /**
     * When given a node representing a bookmark folder, return the <array>...</array> node that is used
     * for adding items to the list.
     * @param parentNode
     * @return
     */
    private static Element getChildrenArray(Node parentNode) throws TransformerException, UnexpectedXmlFileFormatException {
        // Sanity check!  Is this a folder node?
        Node typeNode = XPathAPI.selectSingleNode(parentNode, "string[child::text()='WebBookmarkTypeList']");
        if (typeNode == null)
            throw new UnexpectedXmlFileFormatException("Attempted to find children of an entry that isn't a folder");
        // Look for the "<key>Children</key>" node
        Node childrenNode = XPathAPI.selectSingleNode(parentNode, "key[child::text()='Children']");
        Node arrayNode = null;
        if (childrenNode != null) {
            // Look for the <array>...</array> node immediately following the "Children" node
            arrayNode = childrenNode.getNextSibling();
            while ((arrayNode != null) && (!arrayNode.getNodeName().equals("array")))
                arrayNode = arrayNode.getNextSibling();
            if (arrayNode == null)
                throw new UnexpectedXmlFileFormatException("Children node did not have a sibling");
            if (!arrayNode.getNodeName().equals("array"))
                throw new UnexpectedXmlFileFormatException("Children node had unexpected sibling (" + arrayNode.getNodeName() + " instead of array)");
        }
        return (Element) arrayNode;
    }
    
    /**
     * Create the DOM tree representation of a single bookmark.  For example:
     * <dict>
     *     <key>URIDictionary</key>
     *     <dict>
     *         <key></key>
     *         <string>http://netninja.com/</string>
     *         <key>lastVisitedDate</key>
     *         <string>92610721.6</string>
     *         <key>title</key>
     *         <string>Netninja</string>
     *     </dict>
     *     <key>URLString</key>
     *     <string>http://netninja.com/</string>
     *     <key>WebBookmarkType</key>
     *     <string>WebBookmarkTypeLeaf</string>
     *     <key>WebBookmarkUUID</key>
     *     <string>F20C3446-A879-11D7-A0BB-000393D55BF6</string>
     * </dict>
     * @param doc
     * @param entry
     * @return
     */
    private static Element createBookmarkNode(Document doc, PostEntry entry) {
        Element dictNode = doc.createElement("dict");
        dictNode.appendChild(createTag(doc, "key", "URIDictionary"));
        Element uriNode = doc.createElement("dict");
        uriNode.appendChild(createTag(doc, "key", ""));
        uriNode.appendChild(createTag(doc, "string", entry.getHref()));
        // ASSUMPTION: assuming lastVisitedDate of zero is valid
        uriNode.appendChild(createTag(doc, "key", "lastVisitedDate"));
        uriNode.appendChild(createTag(doc, "string", "0"));
        uriNode.appendChild(createTag(doc, "key", "title"));
        uriNode.appendChild(createTag(doc, "string", entry.getDescription()));
        dictNode.appendChild(uriNode);
        dictNode.appendChild(createTag(doc, "key", "URLString"));
        dictNode.appendChild(createTag(doc, "string", entry.getHref()));
        dictNode.appendChild(createTag(doc, "key", "WebBookmarkType"));
        dictNode.appendChild(createTag(doc, "string", "WebBookmarkTypeLeaf"));
        dictNode.appendChild(createTag(doc, "key", "WebBookmarkUUID"));
        // ASSUMPTION: assuming GUID can just be randomly generated
        dictNode.appendChild(createTag(doc, "string", generateGuid(entry)));
        return dictNode;
    }
    
    /**
     * Create a folder node, suitable for appending to the <key>Children</key><array> ... </array> list
     * @param doc
     * @param name
     * @return
     */
    private static Element createFolderNode(Document doc, String name) {
        Element dictNode = doc.createElement("dict");
        dictNode.appendChild(createTag(doc, "key", "Children"));
        dictNode.appendChild(doc.createElement("array"));
        dictNode.appendChild(createTag(doc, "key", "Title"));
        dictNode.appendChild(createTag(doc, "string", name));
        dictNode.appendChild(createTag(doc, "key", "WebBookmarkType"));
        dictNode.appendChild(createTag(doc, "string", "WebBookmarkTypeList"));
        dictNode.appendChild(createTag(doc, "key", "WebBookmarkUUID"));
        dictNode.appendChild(createTag(doc, "string", generateGuid(name)));
        return dictNode;
    }
    
    /**
     * Generate a GUID for a given entry.  This could be a piece of data inside the entry
     * record, like a GUID string or integer primary key.  It could also be a piece of
     * data generated from the content, like a hash of the URL.  It pretty much only needs
     * to be unique within this test run, since the previous run's entries (and hence their
     * GUIDs) are pruned from the tree before grafting on the new entries. 
     * @param entry
     * @return
     */
    private static String generateGuid(PostEntry entry) {
        return new UID().toString();
    }
    
    private static String generateGuid(String name) {
        return new UID().toString();
    }
    
    /**
     * Create a DOM element that equates to "<name>value</name>"
     * @param name
     * @param value
     * @return
     */
    private static Element createTag(Document doc, String name, String value) {
        Element result = doc.createElement(name);
        result.appendChild(doc.createTextNode(value));
        return result;
    }
    
    /**
     * Locate and return the bookmark bar node in the tree
     * @param doc
     * @return
     */
    public static Element locateBookmarkBarNode(Document doc) throws TransformerException, UnexpectedXmlFileFormatException {
        NodeList nodes = XPathAPI.selectNodeList(doc, "//string[child::text()='BookmarksBar']");
        if (nodes.getLength() == 0) throw new UnexpectedXmlFileFormatException("Could not locate BookmarksBar node");
        if (nodes.getLength() > 1) throw new UnexpectedXmlFileFormatException("Found more than one BookmarksBar node");
        return (Element) nodes.item(0).getParentNode(); 
    }
    
    /** 
     * Given a folder node (typically the bookmark bar), locate an existing del.icio.us node, or create a
     * new one if it does not exist.
     * @param folder
     * @return
     * @throws TransformerException
     * @throws UnexpectedXmlFileFormatException
     */
    public static Element locateCreateSubFolder(Element folder, String name) throws TransformerException, UnexpectedXmlFileFormatException {
        Element children = getChildrenArray(folder);
        if (children == null) throw new UnexpectedXmlFileFormatException("Could not find children node");
        Node foundTitle = XPathAPI.selectSingleNode(children, "dict/string[child::text()='" + name + "']");
        Element result = null;
        // If not found, create; if found, return
        if (foundTitle == null) {
            result = createFolderNode(folder.getOwnerDocument(), name);
            children.appendChild(result);
        }else{
            result = (Element) foundTitle.getParentNode();
        }
        return result;
    }
    
    public String getBookmarkFilename() {
        return this.bookmarkFile;
    }
    
    public int getBackupCount() {
        int result = 0;
        for (int i=1; i<= 9; i++) {
            try{
                File f = new File(this.bookmarkFile + "." + i);
                if (f.exists())
                    result++;
            }catch(Exception e){}
        }
        return result;
    }
    
    public static final void main(String[] argv) throws Exception {
        SafariBookmarkEditor editor = new SafariBookmarkEditor();
        Document doc = editor.load();
        Element bookmarkBarNode = editor.locateBookmarkBarNode(doc);
        Element deliciousFolder = editor.locateCreateSubFolder(bookmarkBarNode, "del.icio.us");
        editor.deleteChildren(deliciousFolder);
        PostEntry[] entries = new PostEntry[1];
        entries[0] = new PostEntry();
        entries[0].setHref("http://test.com");
        entries[0].setDescription("test bookmark");
        entries[0].setTags("TestCategory");
        entries[0].setUser("TestName");
        editor.graftBookmarks(deliciousFolder, "TestName", "TestCategory", entries);
        editor.save("/tmp/Bookmarks.plist", doc);
    }
}
